/*
 * Copyright 2007 Gentoo Foundation
 * Distributed under the terms of the GNU General Public License v2
 * Author: Michael Haubenwallner <haubi@gentoo.org>
 */

#include <config.h>

#include "hpuxplugin.h"

#include <stdlib.h>
#include <string.h>

#if defined(__cplusplus)
extern "C" {
#else
typedef enum { false = 0, true = 1 } bool;
#endif

typedef enum { envrunpath_no = 0, envrunpath_first = 1, envrunpath_second = 2 } envrunpath_t;

/* HPUX-ld:
 * When there is no explicit runpath on the linker commandline (+b runpath),
 * then all library paths (-L) are recorded as runpath
 */
int hpuxplugin(LdPluginData *data)
{
	int argc = 0;
	int err;

	bool withDefaultRunpath = true;					/* +nodefaultrpath */
	envrunpath_t withEnvRunpath = envrunpath_no;	/* +s */

	StringList *tmpArgList = NULL;
	StringList *runpathList = NULL;
	StringList *defaultRunpathList = NULL;
	String *newString = NULL;
	String const *argString;
	char const *argBuffer;
	int argBufferLength;

	do {	/* dummy loop */
		err = -1;

		tmpArgList = StringListCreate(NULL, 0, 0);
		if (tmpArgList == NULL)
			break;

		runpathList = StringListCreate(NULL, 0, 0);
		if (runpathList == NULL)
			break;

		defaultRunpathList = StringListCreate(NULL, 0, 0);
		if (defaultRunpathList == NULL)
			break;

		for(argc = 1; argc < StringListGetSize(data->in->argList); argc++) {
			argString = StringListGetString(data->in->argList, argc);
			argBuffer = StringGetBuffer(argString);
			argBufferLength = StringGetLength(argString);

			if (strcmp(argBuffer, "+s") == 0) {
				if (StringListGetSize(runpathList) == 0)
					withEnvRunpath = envrunpath_first;
				else
					withEnvRunpath = envrunpath_second;
				continue;
			} else
			if (strcmp(argBuffer, "+nodefaultrpath") == 0) {
				withDefaultRunpath = false;
				continue;
#if defined(__ia64)
			} else
			if (strcmp(argBuffer, "+defaultrpath") == 0) {
				withDefaultRunpath = true;
				continue;
#endif
			} else
			if (strncmp(argBuffer, "+b", 2) == 0) {
				/* collect runpath from "+b runpath1:runpathN" */
				char const *curr, *next;

				argBuffer += 2;
				argBufferLength -= 2;

				if (*argBuffer == 0 && argc < StringListGetSize(data->in->argList)) {
					argc++;
					argString = StringListGetString(data->in->argList, argc);
					argBuffer = StringGetBuffer(argString);
					argBufferLength = StringGetLength(argString);
				}

				for(curr = next = argBuffer; *next != '\0'; curr = next+1) {
					for(next = curr; *next != '\0' && *next != ':'; next++);

					if (next - curr <= 1) {
						/* skip empty path */
						continue;
					}

					if (StringListContains(data->in->sysLibpath, curr, next - curr)) {
						/* sys runpath will be added later */
						continue;
					}

					if (StringListContains(runpathList, curr, next - curr)) {
						/* already recorded user runpath */
						continue;
					}

					/* record this path into libpath list */
					if (StringListAppendConcat(runpathList, curr, next - curr, NULL) < 0)
						break;
				}

				/* end "+b" handling */
				continue;
			} else
			if (strncmp(argBuffer, "-L", 2) == 0) {
				/* drop -L arguments from commandline,
				 * we need to pass them before any libs.
				 * The path list already is collected in userLibpath.
				 */
				continue;
			} else
			if (strncmp(argBuffer, "-R", 2) == 0) {
				/* extension:
				 * collect additional default runpath from "-R runpath"
				 */

				argBuffer += 2;
				argBufferLength -= 2;

				if (*argBuffer == 0 && argc < StringListGetSize(data->in->argList)) {
					argc++;
					argString = StringListGetString(data->in->argList, argc);
					argBuffer = StringGetBuffer(argString);
					argBufferLength = StringGetLength(argString);
				}
				if (StringListContains(data->in->sysLibpath, argBuffer, argBufferLength))
					continue;

				if (StringListAppendConcat(defaultRunpathList, argBuffer, argBufferLength, NULL) < 0)
					break;

				continue;
			}

			/* keep other arguments on commandline */
			if (StringListAppendString(tmpArgList, argString) < 0)
				break;
		}
		if (argc < StringListGetSize(data->in->argList)) {
			/* error during argument parsing */
			break;
		}

		/*
		 * first, put libpath list (-L) on the commandline
		 */
		if (StringListAppendListModify(data->out->argList, data->in->userLibpath, 0, -1, "-L", 2, NULL, 0) < 0)
			break;

		/* also put sys libpath list as libpath (-L) on commandline */
		if (StringListAppendListModify(data->out->argList, data->in->sysLibpath, 0, -1, "-L", 2, NULL, 0) < 0)
			break;

		/* do we need to use env-runpath first? */
		if (withEnvRunpath == envrunpath_first
		 && StringListAppendConcat(data->out->argList, "+s", 2, NULL) < 0
		) break;

		/*
		 * now put runpath list (from -R, +b) on the commandline
		 */

		/* append "+b" to commandline */
		if (StringListAppendConcat(data->out->argList, "+b", 2, NULL) < 0)
			break;

		/* if there is no runpath (+b), use libpath as runpath
		 * if not disabled with "+nodefaultrpath"
		 */
		if (StringListGetSize(runpathList) == 0
		 && withDefaultRunpath == true
		 && StringListAppendList(runpathList, data->in->userLibpath, 0, -1) < 0
		) break;

		/* append default runpath list (-R) to runpath list */
		if (StringListAppendList(runpathList, defaultRunpathList, 0, -1) < 0)
			break;

		/* append sys libpath list to runpath list */
		if (StringListAppendList(runpathList, data->in->sysRunpath, 0, -1) < 0)
			break;

		/* create runpath string: runpath1:runpathN */
		newString = StringListJoin(runpathList, NULL, 0, ":", 1, NULL, 0);
		if (newString == NULL)
			break;

		/* append runpath string to commandline */
		if (StringListAppendString(data->out->argList, newString) < 0)
			break;

		/* do we need to use env-runpath second? */
		if (withEnvRunpath == envrunpath_second
		 && StringListAppendConcat(data->out->argList, "+s", 2, NULL) < 0
		) break;

		/*
		 * third, put other arguments on commandline
		 */
		if (StringListAppendList(data->out->argList, tmpArgList, 0, -1) < 0)
			break;

		err = 0;
	} while(0);	/* end dummy loop */

	newString = StringDestroy(newString);
	defaultRunpathList = StringListDestroy(defaultRunpathList);
	runpathList = StringListDestroy(runpathList);
	tmpArgList = StringListDestroy(tmpArgList);

	return err;
}

#if defined(__cplusplus)
}
#endif
